# --------------------------------------------------------------------------------- #
# SIZERHACK wxPython IMPLEMENTATION
#
# Andrea Gavana, @ 12 Jun 2007
# Latest Revision: 12 Jun 2007, 22.00 CET
#
#
# TODO List
#
# 1. Adding more sizers (i.e., wx.FlexGridSizer, wx.GridBagSizer, wx.StaticBoxSizer
#    should not be a big problem. The usefulness of those sizers inside a learning
#    application line SizerHack is still debatable, in my opinion.
#
# 2. Provide an alternative implementation which would use FloatCanvas or OGL,
#    allowing the user to create the sizers-widgets hierarchy using shapes and lines
#    connecting them. The main issue here is that no OGL nor FloatCanvas provide
#    a way to edit a text inside a shape, which is needed to assign sizers/widgets
#    proportions, flags and so on.
#
#
# For All Kind Of Problems, Requests Of Enhancements And Bug Reports, Please
# Write To Me At:
#
# gavana@kpo.kz
# andrea.gavana@gmail.com
#
# Or, Obviously, To The wxPython Mailing List!!!
#
#
# End Of Comments
# --------------------------------------------------------------------------------- #


"""
Description
===========

SizerHack is a small application I wrote while investigating random generated
interfaces. In my opinion, it could be useful to teach wxPython newbies how to
deal with sizers and interfaces layouts in a more "visual" way.


Basic Concepts
==============

Hereafter I present some features of SizerHack, explaining the terminology I use
later on in this part of the documentation.

* Layout Generation: SizerHack generates randomly a layout in a separate frame. Using
  the nesting level you provide (see next point), it creates a layout where widgets
  types are chosen randomly from a stock of 10 different standard wxPython controls,
  added to their sizers using a randomly selected proportion (either 0 or 1) and a
  randomly selected flag (either 0 or wx.EXPAND).
  Selecting a nesting level greater than 1 (which is the default), SizerHack will
  generate a random layout in which sizers are nested.
  In order to avoid a too big separate frame (in term of wx.Size), the proportion
  keyword has a probability distribution of 80% for 0 and 20% for 1 (which means 80%
  of the times a widget or a sizer will be added to their containing sizers with a
  proportion of 0).
  In order to generate a new layout, just select the menu "Edit" => "Create New Layout"
  or hit "Ctrl+L".
  
* Nesting Levels: SizerHack allows the user to select the nesting level for sizers:
  this means the randomly generated layout in the separate frame will have sizers
  nested in sizers nested in sizers... and so on, depending on the nesting level you
  have chosen.
  In order to change the nesting level, just select the menu "Edit" => "Set Nesting Level"
  or hit "Ctrl+N".
  
* Sizer Exposure: there are some cases where being able to see the sizer is helpful,
  such as at design time. SizerHack allows you to see the sizers/widgets hierarchy on
  screen, by painting the sizers/widgets boundaries. Just select the menu "Hints" =>
  "Draw Sizers" or hit "Ctrl+D" to see them. Sizers with the wx.HORIZONTAL orientation
  are painted in solid BLUE, while their children widgets are surrounded by a dotted
  blue line. Sizers with the wx.VERTICAL orientation are in RED.


Description Of Sample Usage
===========================

Once launched, SizerHack will show you a randomly generated layout in a new separate
frame (with wx.STAY_ON_TOP flag). By default, SizerHack creates the initial layout
using a nesting level of 1. You can change this value by choosing the menu "Edit" =>
"Set Nesting Level" or hit "Ctrl+N".

In the SizerHack TreeList control, you will see 2 items, namely the start of the
layout hierarchy in the separate frame. These 2 items represent a wx.Frame and a wx.Panel,
which are fixed for every layout SizerHack generates.
Now, the panel has always a sizer associated with it: what you should do, is right-click
on the wxPanel item and choose "Add Sizer As Child" from the popup menu.

A new line in the TreeList control appears: it contains 3 wx.Choices, asking you to choose
from them: these wx.Choices represent the sizer orientation, proportion and flag that
SizerHack has used to assign this sizer to wxPanel.
If you choose the correct answer from the wx.Choices, the icon next to them will be a
green "Ok" bitmap, otherwise it will be a red "Wrong" cross.

Now that you have added the main sizer, by looking at the separate frame and playing with
it (resize it to see), you should decide which widgets (or sub-sizers) have been added
to the main sizer. If you think a widget is a direct child of the main sizer, simply Drag
and Drop the widget from the separate frame to the item that represents the main sizer
in the TreeList control. If your answer is correct, a new line appears in the TreeList
control, asking you to fill 2 wx.Choices representing the proportion and flag used to
add this window to the main sizer.

In case of nested sizers, once you decide a sub-sizer is a child of the main sizer (or a
sub-sub-sizer is a child of a sub-sizer), just right click on the main sizer (or on the
sub-sizer) and choose "Add Sizer as Child" from the popup menu.

If you are lost, or you don't recognize at a glance which kind of layout you are presented,
you have 2 alternatives:

1) Select the menu "Hints" => "Draw Sizers" to see on screen how the sizers and widgets
   are arranged;
2) Right click on any item in the TreeList control (except the wxFrame) and, from the popup
   menu, choose "Give Help!". SizerHack will either uncover one child of the selected item
   or will fill one of the wx.Choices for you.

Once all the hierarchy is complete and the wx.Choices are filled correctly, SizerHack will
congratulate with you for your wx.SizerSkills!


Supported Platforms
===================

SizerHack has been tested on the following platforms:
  * Windows (Windows XP);


Latest Revision: Andrea Gavana @ 12 Jun 2007, 22.00 CET
Version 0.1

"""

# Version Info
__version__ = "0.1"

import wx
import os
import random
import glob
import weakref
import cPickle

import HyperTreeList as HTL

_sizerDirections = [wx.VERTICAL, wx.HORIZONTAL]
_sizerProportions = [0, 0, 1, 0, 0]
_sizerFlags = [0, wx.EXPAND]

#----------------------------------------------------------------------
def GetMondrianData():
    return \
'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00 \x00\x00\x00 \x08\x06\x00\
\x00\x00szz\xf4\x00\x00\x00\x04sBIT\x08\x08\x08\x08|\x08d\x88\x00\x00\x00qID\
ATX\x85\xed\xd6;\n\x800\x10E\xd1{\xc5\x8d\xb9r\x97\x16\x0b\xad$\x8a\x82:\x16\
o\xda\x84pB2\x1f\x81Fa\x8c\x9c\x08\x04Z{\xcf\xa72\xbcv\xfa\xc5\x08 \x80r\x80\
\xfc\xa2\x0e\x1c\xe4\xba\xfaX\x1d\xd0\xde]S\x07\x02\xd8>\xe1wa-`\x9fQ\xe9\
\x86\x01\x04\x10\x00\\(Dk\x1b-\x04\xdc\x1d\x07\x14\x98;\x0bS\x7f\x7f\xf9\x13\
\x04\x10@\xf9X\xbe\x00\xc9 \x14K\xc1<={\x00\x00\x00\x00IEND\xaeB`\x82' 


def GetMondrianBitmap():
    return wx.BitmapFromImage(GetMondrianImage())


def GetMondrianImage():
    import cStringIO
    stream = cStringIO.StringIO(GetMondrianData())
    return wx.ImageFromStream(stream)


def GetMondrianIcon():
    icon = wx.EmptyIcon()
    icon.CopyFromBitmap(GetMondrianBitmap())
    return icon


# ---------------------------------------------------------------------------- #
# Class SizerDropSource
# Gives Some Custom UI Feedback during the DnD Operations
# ---------------------------------------------------------------------------- #

class SizerDropSource(wx.DropSource):
    """
    Give some custom UI feedback during the drag and drop operation in this
    function. It is called on each mouse move, so your implementation must
    not be too slow.
    """
    
    def __init__(self, win):
        """ Default class constructor. Used internally. """
        
        wx.DropSource.__init__(self, win)
        self._win = win


# ---------------------------------------------------------------------------- #
# Class SizerDragInfo
# ---------------------------------------------------------------------------- #

class SizerDragInfo:

    _map = weakref.WeakValueDictionary()

    def __init__(self, win):
        """ Default class constructor. """
        
        self._id = id(win)
        SizerDragInfo._map[self._id] = win


    def GetWindow(self):
        """ Returns the widget that is being dragged. """
        
        return SizerDragInfo._map.get(self._id, None)


# ---------------------------------------------------------------------------- #
# Class SizerDropTarget
# ---------------------------------------------------------------------------- #

class SizerDropTarget(wx.DropTarget):

    def __init__(self, parent):
        """ Default class constructor. """
        
        wx.DropTarget.__init__(self)

        self._parent = parent
        self._dataobject = wx.CustomDataObject(wx.CustomDataFormat("SizerHack"))
        self.SetDataObject(self._dataobject)


    def OnData(self, x, y, dragres):
        """ Handles the OnData() method to call the real DnD routine. """
        
        if not self.GetData():
            return wx.DragNone

        draginfo = self._dataobject.GetData()
        drginfo = cPickle.loads(draginfo)
        
        return self._parent.OnDropTarget(x, y, drginfo.GetWindow())


class SizerPanel(wx.Panel):

    def __init__(self, parent, inputWidget, treeItem, column):

        wx.Panel.__init__(self, parent, style=wx.TAB_TRAVERSAL)

        self.inputWidget = inputWidget
        self.treeItem = treeItem
        
        self.SetBackgroundColour(parent.GetBackgroundColour())        
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.bitmap = wx.StaticBitmap(self, -1, wx.Bitmap("images/question.png", wx.BITMAP_TYPE_PNG))
        
        if column == 1:     # Sizer orientation
            choices = ["wx.VERTICAL", "wx.HORIZONTAL"]
            eventHandler = self.OnSizerOrientation
        elif column == 2:   # Proportions
            choices=["0", "1"]
            eventHandler = self.OnProportions
        else:               # Flags
            choices=["0", "wx.EXPAND"]
            eventHandler = self.OnFlags

        self.choice = wx.Choice(self, -1, choices=choices)
        sizer.Add(self.choice, 0, wx.EXPAND|wx.ALIGN_CENTER_VERTICAL|wx.TOP|wx.BOTTOM, 2)
        sizer.Add(self.bitmap, 0, wx.EXPAND|wx.ALIGN_CENTER_VERTICAL|wx.LEFT|wx.RIGHT, 5)
        self.SetSizer(sizer)
        sizer.Layout()
        self.Fit()

        self.choice.Bind(wx.EVT_CHOICE, eventHandler)


    def OnSizerOrientation(self, event):

        userChoice = eval(event.GetString())
        self.SetAnswerBitmap(userChoice == self.inputWidget.GetOrientation(), 0)


    def OnProportions(self, event):

        userChoice = eval(event.GetString())
        proportion = self.inputWidget.GetProportion()
        proportion = (proportion == -1 and [1] or [proportion])[0]
        self.SetAnswerBitmap(userChoice == proportion, 1)


    def OnFlags(self, event):

        userChoice = eval(event.GetString())
        widgetFlags = self.inputWidget.GetFlag()
        widgetFlags &= ~wx.ALL
        widgetFlags = (widgetFlags < 0 and [wx.EXPAND] or [widgetFlags])[0]
        self.SetAnswerBitmap(userChoice == widgetFlags, 2)

        
    def SetAnswerBitmap(self, correctChoice, question):
        
        if correctChoice:
            self.bitmap.SetBitmap(wx.Bitmap("images/ok.png", wx.BITMAP_TYPE_PNG))
        else:
            self.bitmap.SetBitmap(wx.Bitmap("images/error.png", wx.BITMAP_TYPE_PNG))

        self.inputWidget.SetAnswers(question, correctChoice)
        
        self.GetSizer().Layout()
        self.bitmap.Refresh()
        self.GetParent().RefreshLine(self.treeItem)
        self.GetGrandParent().ExerciseFinished()


    def GiveHelp(self, helpLevel):

        if helpLevel == 1:    # uncover the orientation flag
            orientation = self.inputWidget.GetOrientation()
            self.choice.SetSelection(orientation==wx.VERTICAL)
        
        elif helpLevel == 2:
            proportion = self.inputWidget.GetProportion()
            self.choice.SetSelection(proportion==1)

        else:
            flag = self.inputWidget.GetFlag()
            flag &= ~wx.ALL
            self.choice.SetSelection(flag==wx.EXPAND)
            
        self.SetAnswerBitmap(True, helpLevel-1)        

            
class SizerTreeItemData(object):
    """ Class used to store information about a window or a sizer. """

    def __init__(self, inputWidget, label="", proportion=-1, flag=-1):
        """ Default class constructor. """

        self._window = self._sizer = False

        if isinstance(inputWidget, wx.Window):
            self._window = inputWidget
            self._answers = [1, 0, 0]
            self._helpLevel = 2
            if isinstance(inputWidget, wx.Panel):
                self._helpLevel = -1
        elif isinstance(inputWidget, wx.Sizer):
            self._sizer = inputWidget
            self._answers = [0, 0, 0]
            self._helpLevel = -4
            
        self._proportion = proportion
        self._flag = flag
        self._label = label
        

    def IsWindow(self):
        """ Returns whether the item is a widget or not. """

        return self._window


    def IsSizer(self):
        """ Returns whether the item is a sizer or not. """

        return self._sizer


    def GetOrientation(self):

        return self._sizer.GetOrientation()


    def GetProportion(self):

        return self._proportion


    def GetFlag(self):

        return self._flag        


    def SetAnswers(self, question, isCorrect):

        self._answers[question] = isCorrect


    def GetAnswers(self):

        return self._answers
    

    def SetHelpLevel(self, level):

        self._helpLevel = level


    def GetHelpLevel(self):

        return self._helpLevel        

        
#----------------------------------------------------------------------------
# There is no direct connection between a sizer and its
# containing sizer (if any). So find it by finding first
# the sizer's parent window. This window must have a 
# sizer and it will be the containing sizer or will
# contain the containing sizer.
#----------------------------------------------------------------------------

    def GetContainingSizer(self, szr):
        """ Utility method. """

        parent = self.FindSizerParentWindow(szr)
        containingSizer = None
        if parent and parent.GetSizer():
            containingSizer = self.FindParentSizer(parent.GetSizer(), szr)

        return containingSizer

#----------------------------------------------------------------------------
# Check ancestor and ancestor's descendent sizers to find
# the sizer containing szr, i.e., the parent sizer for szr.
#----------------------------------------------------------------------------

    def FindParentSizer(self, ancestor, szr):
        """ Utility method. """

        children = ancestor.GetChildren()
        
        for child in children:
            if child.IsSizer():
                if szr == child.GetSizer():
                    return ancestor

            # Recurse.
            s = self.FindParentSizer(child.GetSizer(), szr)
            if s:
                return s

        return None


    def FindSizerParentWindow(self, szr):
        """ Utility method. """

        children = szr.GetChildren()
        
        for child in children:
            if child.IsWindow():
                return child.GetWindow().GetParent()

            elif child.IsSizer():
              # Recurse.
              return self.FindSizerParentWindow(child.GetSizer())
  
        return None


    def Flash(self, dc):

        if self._window:

            x1, y1 = self._window.GetPosition()
            size = self._window.GetSize()
            sizer = self._window.GetContainingSizer()
            if not sizer:
                direction = wx.VERTICAL
            else:
                direction = sizer.GetOrientation()
            size = wx.Size(size.x+1, size.y+1)
        else:

            x1, y1 = self._sizer.GetPosition()
            size = self._sizer.GetSize()
            direction = self._sizer.GetOrientation()

        self.DrawOnDC(dc, size, x1, y1, direction)


    def DrawOnDC(self, dc, size, x1, y1, direction):
        """ Actually draws on the wx.ScreenDC. """

        dc.SetBrush(wx.TRANSPARENT_BRUSH)

        pen = wx.Pen(wx.BLUE, 2)
        if direction == wx.VERTICAL:
            pen.SetColour(wx.RED)
        if self.IsWindow():
            pen.SetStyle(wx.SHORT_DASH)
            
        dc.SetPen(pen)
        dc.DrawRectangle(x1, y1, size.GetWidth(), size.GetHeight())
        

class SizerTree(HTL.HyperTreeList):
    """ The HyperTreeList used to display the window/sizer hierarchy. """

    def __init__(self, parent, id=wx.ID_ANY, pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.SUNKEN_BORDER|wx.TR_HIDE_ROOT|
                 wx.TR_NO_BUTTONS|wx.TR_HAS_VARIABLE_ROW_HEIGHT|wx.TR_ROW_LINES|wx.gizmos.TR_COLUMN_LINES):
        """ Default class constructor. """
        
        HTL.HyperTreeList.__init__(self, parent, id, pos, size, style)

        self.mainFrame = wx.GetTopLevelParent(self)
        self._filenames = []
        self.CreateImageList()

        self._dropTarget = SizerDropTarget(self)
        self.GetMainWindow().SetDropTarget(self._dropTarget)

        self.AddColumn("Widgets & Sizers")
        self.AddColumn("Orientation")
        self.AddColumn("Proportion")
        self.AddColumn("Flag")
        self.SetMainColumn(0) # the one with the tree in it...
        self.SetColumnWidth(0, 175)
        
        self.GetMainWindow().Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)
        self.GetMainWindow().Bind(wx.EVT_RIGHT_UP, self.OnRightUp)
        

    def CreateImageList(self):
        """ Creates the wx.TreeCtrl image list. """
        
        ilist = wx.ImageList(24, 24)
        images = glob.glob("images/wx*.png")
        for img in images:
            ilist.Add(wx.Bitmap(img, wx.BITMAP_TYPE_PNG))
            self._filenames.append(os.path.split(os.path.splitext(img)[0])[1])

        self.AssignImageList(ilist)
        

    def GetSuitableImage(self, win):
        """ Returns a suitable bitmap depending on the widget type. """

        found = 0
        winclassname = repr(win).split(";")[0]
        winclassname = winclassname.strip(">").strip("<")
        winclassname = winclassname.replace("_controls.", "").replace("_windows.", "")
        winclassname = winclassname.replace("_stc.", "").replace("_core.", "")
        winclassname = winclassname.replace("_animate.", "").replace("_wizard.", "")
        winclassname = winclassname.replace("_calendar.", "")
        winclassname = winclassname.replace("Py", "")

        if "__main__" in winclassname:
            winclassname = win.GetClassName()
            winclassname = winclassname[0:2] + "." + winclassname[2:]
        
        for count, files in enumerate(self._filenames):
            if winclassname == files: 
                found = 1
                break

        if found:
            return count

        return self.GetImageList().GetImageCount() - 2


    def UpdateSizerHierarchy(self):
        """ Updates the sizer hierarchy. """

        self.Freeze()
        # Delete all the node in the tree.
        self.DeleteAllItems()

        # Create a root node(which we're hiding anyhow)
        rootId = self.AddRoot("Root")

        # Loop over top level windows and build tree for each.
        for item in wx.GetTopLevelWindows():
            # Skip SizerTree frame.
            if isinstance(item, SizerHack) or isinstance(item, wx.TextEntryDialog):
                continue

            self.AddSizerNodes(rootId, item)

        for i in xrange(1, self.GetMainWindow().GetColumnCount()):
            self.SetColumnWidth(i, wx.LIST_AUTOSIZE_USEHEADER)

        self.Thaw()

        
    def AddSizerNodes(self, parent, win, prp=-1, flg=-1):
        """ Adds children nodes in the sizers hierarchy. """

        if isinstance(win, wx.Sizer):
            self.AddSizerNodes2(parent, win, prp, flg)
            return
        
        label = win.GetClassName()

        # Add title for frames and labels for wxControls
        if isinstance(win, wx.Frame):
            label += " (" + win.GetTitle() + ")"

        elif isinstance(win, wx.Control):
            text = win.GetLabel()
            if len(text) > 30:
                text = text[0:30] + "..."
            if text:
                label += " (" + text + ")"

        image = self.GetSuitableImage(win)
        data = SizerTreeItemData(win, label, prp, flg)
        leaf = self.AppendItem(parent, label, image=image)
        self.SetPyData(leaf, data)
        
        if not isinstance(win, wx.Panel) and not isinstance(win, wx.Frame):
            self.HideItem(leaf)
            self.SetItemText(leaf, 1, "None")
            for column in xrange(2, 4):
                wnd = SizerPanel(self.GetMainWindow(), data, leaf, column)
                self.SetItemWindow(leaf, wnd, column)

        sizer = win.GetSizer()

        if sizer:
            self.AddSizerNodes(leaf, sizer)
        else:
            children = win.GetChildren()
            for child in children:
                if not isinstance(child, wx.Frame):
                    # Recurse.
                    self.AddSizerNodes(leaf, child)


    def AddSizerNodes2(self, parent, sizer, prp=-1, flg=-1):
        """ Adds children nodes in the sizers hierarchy. """

        label = sizer.GetClassName()
        image = self.GetSuitableImage(sizer)
                
        # Add leaf to tree.
        data = SizerTreeItemData(sizer, label, prp, flg)
        leaf = self.AppendItem(parent, label, image=image)
        for column in xrange(1, 4):
            wnd = SizerPanel(self.GetMainWindow(), data, leaf, column)
            self.SetItemWindow(leaf, wnd, column)

        self.SetPyData(leaf, data)
        self.HideItem(leaf)
        
        children = sizer.GetChildren()
        
        for child in children:
            if child.IsWindow():
                # Recurse.
                self.AddSizerNodes(leaf, child.GetWindow(), child.GetProportion(),
                                   child.GetFlag())
            elif child.IsSizer():
                # Recurse.
                self.AddSizerNodes(leaf, child.GetSizer(), child.GetProportion(),
                                   child.GetFlag())


    def ExpandTree(self, item):
        """ Expands all the children of item. """
        
        if item.IsHidden():
            return

        if not self.CheckUserSolution(item):
            return
        
        if item != self.GetRootItem():
            self.Expand(item)
            
        child, cookie = self.GetFirstChild(item)
        while child:
            self.ExpandTree(child)
            child, cookie = self.GetNextChild(item, cookie)


    def CheckUserSolution(self, item):

        child, cookie = self.GetFirstChild(item)
        while child:
            if child.IsHidden():
                return False
            child, cookie = self.GetNextChild(item, cookie)

        return True        


    def DrawSizersOutline(self, dc, item=None):
        """
        There are some cases where being able to see the sizer is helpful, such
        as at design time. This method can be called to see the outline; it needs
        to be called whenever the containing window is resized or repainted.
        """

        if item is None:
            item, cookie = self.GetFirstChild(self.GetRootItem())
            item, cookie = self.GetFirstChild(item)
            
        child, cookie = self.GetFirstChild(item)
        while child:
            data = self.GetPyData(child)
            data.Flash(dc)
            self.DrawSizersOutline(dc, child)
            child, cookie = self.GetNextChild(item, cookie)
            

    def ExerciseFinished(self):

        if not self.WalkAllItems():
            return

        msg = "Congratulations! You have completed the SizerHack game correctly!"
        self.mainFrame.RunError(msg, "Congratulations!", "Information")


    def WalkAllItems(self, item=None):
        
        if item is None:
            item, cookie = self.GetFirstChild(self.GetRootItem())
            item, cookie = self.GetFirstChild(item)
        
        child, cookie = self.GetFirstChild(item)
        while child:
            data = self.GetPyData(child)
            if not all(data.GetAnswers()):
                return False

            if not self.WalkAllItems(child):
                return False
            child, cookie = self.GetNextChild(item, cookie)

        return True


    def OnDropTarget(self, x, y, dropWindow):

        item, flags, column = self.HitTest(wx.Point(x, y))
        if not item:
            return wx.DragCancel

        if item.IsHidden():
            return wx.DragCancel
        
        found = 0
        child, cookie = self.GetFirstChild(item)
        while child:
            data = self.GetPyData(child)
            if data._window == dropWindow and child.IsHidden():
                found = 1
                break
            child, cookie = self.GetNextChild(item, cookie)

        if not found:
            msg = "Attention! The window you have dragged:\n\n%s\n\nis not a direct child" \
                  "of the drop item you chose:\n\n%s\n\nPlease look carefully at the layout!" \
                  %(dropWindow.GetClassName(), self.GetItemText(item))
            self.mainFrame.RunError(msg, "Attention!", "Attention")
            return wx.DragCancel
        
        parentItem = self.GetItemParent(child)
        self.HideItem(child, False)
        self.Expand(parentItem)

        for i in xrange(1, self.GetMainWindow().GetColumnCount()):
            self.SetColumnWidth(i, wx.LIST_AUTOSIZE)
        
        return wx.DragMove


    def OnRightDown(self, event):
        
        try:
            pt = event.GetPosition()
        except:
            pt = event.GetPoint()
            
        item, flags, column = self.HitTest(pt)
        
        if item:
            self.item = item
            self.SelectItem(item)
        else:
            self.item = None


    def OnRightUp(self, event):

        item = self.item
        
        if not item:
            event.Skip()
            return
        
        if item.IsHidden():
            event.Skip()
            return

        data = self.GetPyData(item)
        if isinstance(data._window, wx.Frame):
            event.Skip()
            return

        menu = wx.Menu()
        if isinstance(data._window, wx.Panel) or isinstance(data._sizer, wx.Sizer):
            item = wx.MenuItem(menu, wx.ID_ANY, "Add Sizer As Child")
            item.SetBitmap(self.mainFrame.CreateBitmap("childsizer"))
            menu.AppendItem(item)
            self.Bind(wx.EVT_MENU, self.OnAddSizer, item)

        item = wx.MenuItem(menu, wx.ID_ANY, "Give Help!")
        item.SetBitmap(self.mainFrame.CreateBitmap("help"))
        menu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnHelp, item)
            
        self.PopupMenu(menu)
        menu.Destroy()


    def OnAddSizer(self, event):

        children = self.item.GetChildren()
        hasSizers = 0
        for child in children:
            data = self.GetPyData(child)
            if data._sizer and child.IsHidden():
                hasSizers = 1
                break

        if not hasSizers:
            msg = "Attention! The item:\n\n%s\n\nhas no (other) sizers as children.\n" \
                  "Please examine carefully the layout!"%self.GetItemText(self.item)
            self.mainFrame.RunError(msg, "Attention!", "Attention")
            return

        parentItem = self.GetItemParent(child)
        self.HideItem(child, False)
        self.Expand(parentItem)

        for i in xrange(1, self.GetMainWindow().GetColumnCount()):
            self.SetColumnWidth(i, wx.LIST_AUTOSIZE)        
        

    def OnHelp(self, event):

        data = self.GetPyData(self.item)
        helpLevel = data.GetHelpLevel()
        if helpLevel == -1:       # it's the main panel
            # We uncover one of the item if it's hidden
            self.UncoverHiddenItem(self.item)
            return
        elif helpLevel == 0:      # it's a sizer
            # We uncover one of the item if it's hidden
            self.UncoverHiddenItem(self.item)
        elif helpLevel == 4 and data.IsWindow():
            msg = "SizerHack can not provide any more help on this item!"
            self.mainFrame.RunError(msg, "All fields filled", "Information")
            return
        elif helpLevel < 0:
            self.GetItemWindow(self.item, helpLevel+5).GiveHelp(helpLevel+5)
        else:
            self.GetItemWindow(self.item, helpLevel).GiveHelp(helpLevel)

        data.SetHelpLevel(helpLevel+1)
        self.SetPyData(self.item, data)
            
        event.Skip()


    def UncoverHiddenItem(self, item):

        child, cookie = self.GetFirstChild(item)
        while child:
            if child.IsHidden():
                self.HideItem(child, False)
                self.Expand(item)
                for i in xrange(1, self.GetMainWindow().GetColumnCount()):
                    self.SetColumnWidth(i, wx.LIST_AUTOSIZE)
                return
            child, cookie = self.GetNextChild(item, cookie)

        msg = "SizerHack can not provide any more help on this item!"
        self.mainFrame.RunError(msg, "Sub-Layout completed", "Information")

    
class RandomFrame(wx.Frame):

    def __init__(self, parent, id=wx.ID_ANY, title="", pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.DEFAULT_FRAME_STYLE|wx.STAY_ON_TOP,
                 nestingLevels=None):

        wx.Frame.__init__(self, parent, id, title, pos, size, style)

        self.mainFrame = parent        
        self.RandomLayout(nestingLevels)

        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.Bind(wx.EVT_SIZE, self.OnSize)

        self.SetIcon(GetMondrianIcon())        
        self.CenterOnScreen()
        self.Show()


    def RandomLayout(self, nestingLevels):

        self.SetSize((400, 300))
        self.Freeze()
        
        if hasattr(self, "mainPanel"):
            self.mainPanel.Destroy()
            
        self.mainPanel = wx.Panel(self)
        self.sizerHierarchy = {}

        for indx in xrange(nestingLevels):

            sizerDirection = random.choice(_sizerDirections)
            box = wx.BoxSizer(sizerDirection)
            numSlots = random.randint(1, 3)
            boxPosition = random.randint(1, numSlots)
            boxAdded = 0
                       
            for slot in xrange(numSlots+1):

                sizerProportion = random.choice(_sizerProportions)
                sizerFlag = random.choice(_sizerFlags)

                if indx == 0 or slot != boxPosition or boxAdded == 1:
                    item = self.CreateControl()
                else:
                    item = self.sizerHierarchy[indx-1]
                    boxAdded = 1
                    
                box.Add(item, sizerProportion, sizerFlag | wx.ALL, 5)
                
            self.sizerHierarchy[indx] = box

        mainSizer = self.sizerHierarchy[nestingLevels-1]
        self.mainPanel.SetSizer(mainSizer)
        mainSizer.SetSizeHints(self.mainPanel)
        mainSizer.Layout()
        self.Layout()

        self.Thaw()
        self.SizeLayout()
        

    def CreateControl(self):

        controlChoice = random.randint(1, 10)
        choices = ["zero", "one", "two", "three", "four", "five"]
        randSelection = len(choices) - 1

        if controlChoice == 1:
            widget = wx.Button(self.mainPanel, -1, "Hello!")
        elif controlChoice == 2:
            widget = wx.CheckBox(self.mainPanel, -1, "wxPython Rules")
        elif controlChoice == 3:
            widget = wx.Choice(self.mainPanel, -1, choices=choices)
            widget.SetSelection(random.randint(0, randSelection))
        elif controlChoice == 4:
            widget = wx.Gauge(self.mainPanel, -1, 50)
            widget.SetValue(30)
        elif controlChoice == 5:
            widget = wx.RadioButton(self.mainPanel, -1, "RadioButton")
        elif controlChoice == 6:
            widget = wx.Slider(self.mainPanel, -1, 25, 1, 100,
                               style=wx.SL_HORIZONTAL | wx.SL_AUTOTICKS | wx.SL_LABELS)
            widget.SetTickFreq(5, 1)
        elif controlChoice == 7:
            bmp = wx.Bitmap("images/staticbitmap.bmp", wx.BITMAP_TYPE_BMP)
            widget = wx.StaticBitmap(self.mainPanel, -1, bmp)
        elif controlChoice == 8:
            widget = wx.TextCtrl(self.mainPanel, -1,
                                 "Here is a looooooooooooooong line of text.\n\n"
                                 "The quick brown fox jumped over the lazy dog...",
                                 style=wx.TE_MULTILINE|wx.TE_PROCESS_ENTER)
        elif controlChoice == 9:
            widget = wx.StaticText(self.mainPanel, -1, "I hope you like SizerHack!")
            widget.SetBackgroundColour(wx.NamedColor("yellow"))
        elif controlChoice == 10:
            widget = wx.CheckListBox(self.mainPanel, -1, choices=choices)
            widget.SetSelection(random.randint(0, randSelection))
            widget.Check(random.randint(0, randSelection))

        widget.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        
        return widget


    def Redraw(self):

        self.Refresh()        
        if self.mainFrame.drawSizers:
            dc = wx.ClientDC(self.mainPanel)
            wx.CallAfter(self.mainFrame.sizerTree.DrawSizersOutline, dc)


    def SizeLayout(self):

        size = self.GetSize()
        self.SetSize((size.x+1, size.y+1))
        self.SetSize(size)


    def OnClose(self, event):

        self.mainFrame.randomFrame = None
        self.mainFrame.sizerTree.DeleteAllItems()
        self.Destroy()
        
        
    def OnSize(self, event):

        event.Skip()
        self.Redraw()


    def OnLeftDown(self, event):

        event.Skip()
        widget = event.GetEventObject()
        wx.CallAfter(self.DoDragDrop, widget)


    def DoDragDrop(self, widget):
        
        draginfo = SizerDragInfo(widget)
        drginfo = cPickle.dumps(draginfo)
        dataobject = wx.CustomDataObject(wx.CustomDataFormat("SizerHack"))
        dataobject.SetData(drginfo)
        dragSource = SizerDropSource(self)
        dragSource.SetData(dataobject)
        dragSource.DoDragDrop(wx.Drag_DefaultMove)

                    
class SizerHack(wx.Frame):

    def __init__(self, parent, id=wx.ID_ANY, title="", pos=wx.DefaultPosition,
                 size=wx.DefaultSize, style=wx.DEFAULT_FRAME_STYLE):

        wx.Frame.__init__(self, parent, id, title, pos, size, style)

        self.nestingLevels = 1
        self.iteration = 1
        self.drawSizers = False
        
        self.randomFrame = None

        self.SetIcon(GetMondrianIcon())         
        self.CreateMenuBar()
        self.CreateBar()
        self.CreateSizerTree()
        
        self.SetPosition((5, 5))
        self.Show()

        self.RandomLayout()


    def CreateSizerTree(self):
        
        panel = wx.Panel(self, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)

        # Put the tree in the panel
        self.sizerTree = SizerTree(panel, -1)
        sizer.Add(self.sizerTree, 1, wx.EXPAND)
        panel.SetSizerAndFit(sizer)

        
    def CreateMenuBar(self):
    
        # Prepare the menu bar
        menuBar = wx.MenuBar()

        # 1st menu from left
        fileMenu = wx.Menu()
        item = wx.MenuItem(fileMenu, wx.ID_ANY, "E&xit\tCtrl+Q", "Exit SizerHack")
        item.SetBitmap(self.CreateBitmap("exit"))
        fileMenu.AppendItem(item)
        
        editMenu = wx.Menu()
        item = wx.MenuItem(editMenu, wx.ID_ANY, "Create New &Layout\tCtrl+L", "Generates Randomly A New Layout")
        item.SetBitmap(self.CreateBitmap("layout"))
        editMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnLayout, item)
        item = wx.MenuItem(editMenu, wx.ID_ANY, "Set &Nesting Level\tCtrl+N", "Sets The Sizers' Nesting Level")
        item.SetBitmap(self.CreateBitmap("nesting"))
        editMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnNestingLevel, item)

        hintMenu = wx.Menu()    
        item = wx.MenuItem(hintMenu, wx.ID_ANY, "Draw &Sizer\tCtrl+D", "Sets The Sizers' Nesting Level", kind=wx.ITEM_CHECK)
        hintMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnDrawSizer, item)

        helpMenu = wx.Menu()
        item = wx.MenuItem(helpMenu, wx.ID_ANY, "&About SizerHack...", "About SizerHack")
        item.SetBitmap(self.CreateBitmap("about"))
        helpMenu.AppendItem(item)
        self.Bind(wx.EVT_MENU, self.OnAbout, item)
        
        # Add menu to the menu bar
        menuBar.Append(fileMenu, "&File")
        menuBar.Append(editMenu, "&Edit")
        menuBar.Append(hintMenu, "&Hints")
        menuBar.Append(helpMenu, "H&elp")
        self.SetMenuBar(menuBar)


    def CreateBar(self):

        statusbar = self.CreateStatusBar(2, wx.ST_SIZEGRIP)
        statusbar.SetStatusWidths([-2, -1])
        # statusbar fields
        statusbar_fields = [("SizerHack Demo, Andrea Gavana @ 12 Jun 2007"),
                            ("Welcome To wxPython!")]

        for i in range(len(statusbar_fields)):
            statusbar.SetStatusText(statusbar_fields[i], i)
                

    def CreateBitmap(self, png):

        return wx.Bitmap("images/%s.png"%png, wx.BITMAP_TYPE_PNG)


    def RandomLayout(self):

        wx.BeginBusyCursor()
        
        if not self.randomFrame:
            self.randomFrame = RandomFrame(self, size=(400, 300), nestingLevels=self.nestingLevels)
        else:
            self.randomFrame.RandomLayout(self.nestingLevels)
            
        self.randomFrame.SetTitle("Random Generated Frame #%d"%self.iteration)
        self.UpdateTree()
        self.iteration += 1

        wx.EndBusyCursor()
        

    def UpdateTree(self):
        """ Updates the sizer hierarchy. """

        self.sizerTree.Freeze()
        self.sizerTree.UpdateSizerHierarchy()
        root = self.sizerTree.GetRootItem()
        if root:
            self.sizerTree.ExpandTree(root)
            
        self.sizerTree.Thaw()


    def RunError(self, msg, title, kind):

        style = wx.OK | wx.STAY_ON_TOP
        
        if kind == "Attention":
            style |= wx.ICON_EXCLAMATION 
        else:
            style |= wx.ICON_INFORMATION
            
        dlg = wx.MessageDialog(self, msg, title, style)
        dlg.ShowModal()
        dlg.Destroy()
        

    def OnLayout(self, event):

        self.RandomLayout()


    def OnNestingLevel(self, event):

        dlg = wx.TextEntryDialog(self, "Please Enter The New Nesting Level",
                                 "Nesting Level Dialog", str(self.nestingLevels))

        if dlg.ShowModal() != wx.ID_OK:
            dlg.Destroy()
            return
        
        self.nestingLevels = int(dlg.GetValue())            
        dlg.Destroy()
        wx.SafeYield()
        
        self.RandomLayout()


    def OnDrawSizer(self, event):

        self.drawSizers = event.IsChecked()
        self.randomFrame.SizeLayout()


    def OnAbout(self, event):

        msg = "This is the about dialog of SizerHack.\n\n" + \
        "Version %s"%__version__ + "\n"+ \
        "Author: Andrea Gavana @ 12 Jun 2007\n\n" + \
        "Please report any bug/request of improvements\n" + \
        "to me at the following addresses:\n\n" + \
        "andrea.gavana@gmail.com\ngavana@kpo.kz."
              
        self.RunError(msg, "About SizerHack...", "Information")
        
        
def main():

    app = wx.PySimpleApp()
    frame = SizerHack(None, -1, "SizerHack Demo", size=(600, 500))
    app.MainLoop()


if __name__ == "__main__":
    main()

